/*
program:  for ifproxy (simulating a network card's interface for ECO32)
purpose:  part of ECO32 simulation process, providing just the interface to a virtual network 
          interface card
task:     Developed as a project in course "Mikroprozessortechnik II" (simulation of 
          processors) at the University of Applied Sciences Giessen-Friedberg in 
          winter term 2003. The course was led by Prof. Dr. Hellwig Geisse.
authors:  Jochen Mller
          Mathias Schfer
          Mattias Ansorg <matthias@ansorgs.de>
license:  public domain
date:     2003-11-03 to 
usage:    
*/

#include <stdio.h>
#include <rpc/rpc.h>          /* always needed for RPC */
#include <pthread.h>
#include <sys/types.h>        /* for getpid() */
#include <unistd.h>           /* gethostname() getpid()*/

#include "IfProxy.h"
#include "ViNet_Protocol.h"   /* need this too; will be generated by rpcgen */
#include "IfProxy_Protocol.h" /* need this too; will be generated by rpcgen */

/* how to contact ViNet server: client handle */
CLIENT *cl;
/* what ViNet rpc peer server to contact; needed in error messages throughout the program */
char *server;

/* 
 * dispatch routine, defined in server stub IfProxy_Protocol_svc.c generated by rpcgen
 */
extern void ifendprog_1(struct svc_req *, register SVCXPRT *);

/*
 * implementation of routine provided via rpc
 * returns pointer on int with value 0 on success, value nonzero otherwise
 */
int * printmessage_1_svc(char *msg, struct svc_req *rqstp) {
  static int result; /* must be static! */
  FILE *f;
  
  f = fopen("/dev/stdout", "w");
  if (f == NULL) { 
    result = 1;
    return (&result);
  }
  fprintf(f, "rpc call received: print message \"%s\"\n", msg);
  fclose(f);
  result = 0;
  return &result;
}

/*
 * initialize the rpc server provided by IfProxy
 */
int svc_init (int *ifendprog) {
  register SVCXPRT *transp;
  pid_t pid; /* process id of server thread; used as an identifying offset to IFENDPROG */
   
  /* get my pid */
  pid = getpid();

  *ifendprog = IFENDPROG + (int)pid;
  pmap_unset (*ifendprog, IFENDVERS);

  transp = svcudp_create(RPC_ANYSOCK);
  if (transp == NULL) {
    fprintf (stderr, "cannot create udp service\n");
    return 1;
  }
  if (!svc_register(transp, *ifendprog, IFENDVERS, ifendprog_1, IPPROTO_UDP)) {
    fprintf (stderr, "unable to register (ifendprog, IFENDVERS, udp)\n");
    return 1;
  }
  return 0;
}

/*
 * main function of the rpc server thread
 */
void * svc_thread_main(void * arg) {
  static int result;
  int *result_ptr;
  char hostname[HOSTNAMELEN]; /* name of localhost */
  int ifendprog; /* effectual number for rpc program provided by this server thread */
  int ifendvers = IFENDVERS; /* version of rpc program provided by this server thread */

  /* initialize rpc server; must be done in this thread somewhy or svc_run() won't do*/
  result = svc_init(&ifendprog); /* POST: ifendprog == IFENDPROG + own pid */
  if (result != 0) {
    fprintf(stderr, "Error: rpc server initialisation failed.\n");
    return &result;
  }
  
  /* get name of localhost */
  result = gethostname(hostname, HOSTNAMELEN);
  if (result != 0) {
    fprintf(stderr, "Error: could not get valid hostname, aborting server thread\n");
    result = 1;
    return &result;
  }
  
  /* make this IfProxy rpc server known to ViNet rpc server via rpc call*/
  result_ptr = nicregister_1(hostname, ifendprog, ifendvers, cl);
  if (result_ptr == NULL) {
    /* An error occurred while calling the server. Print error message and die. */
    clnt_perror(cl, server);
    fprintf(stderr, "Error: unable to call register routine at ViNet server %s\n", server);
    result = 1;
    return &result;
  }
  else {
    result = *result_ptr;
  }
  /* Okay, we successfully called the remote procedure. */
  if (result != 0) {
    fprintf(stderr, "Error: register routine call at ViNet server %s returned an error\n",     
      server);    
    result = 1;
    return &result;
  }
  else {
    /* we're now callable from ViNet server */
    printf("rpc called: successfully registered as IfProxy server at ViNet server on %s\n", server);
  }
  
  /* start my own IfProxy rpc server */
  svc_run (); /* never returns if svc_init() correctly done, but if not: */
  result = 1;
  return &result; /* according to man pthread_create(3) equivalent to: pthread_exit(&result); */
}

/*
 * IfProxy main
 * initiates the test calls
 */
int main(int argc, char *argv[]) {
  int result;
  int *result_ptr = &result;
  char *message;  
  pthread_t svc_thread; /* rpc server thread in IfProxy */
  
  if (argc < 3) {
    fprintf(stderr, "usage: %s host message\n", argv[0]);
    exit(1);
  }
  
  /* Save values of command line arguments */ 
  server = argv[1];
  message = argv[2];
 
  /* Create client "handle" used for calling MESSAGEPROG on the ViNet server as designated on 
   * the command line */
  cl = clnt_create(server, NICENDPROG, NICENDVERS, "udp");
  if (cl == NULL) {
    /* Couldn't establish connection with server. Print error message and die. */
    clnt_pcreateerror(server);
    exit(1);
  }
 
  /* create and run server thread */
  result = pthread_create(&svc_thread, NULL, &svc_thread_main, NULL);
  /* first argument must be a _valid_ pointer to pthread_t instance, else you'll get strange 
   * segfaults. The instance is _not_ created in pthread_create()! */
  if (result != 0) {
    fprintf(stderr, "Error: server thread could not be created\n");
  }
  
  /* wait on my IfProxy rpc server being ready to use and known to ViNet rpc server */
  sleep(2); /* to be replaced by condition sync with svc_thread */
  
  /* TEST OF VINET RPC SERVER AND ASSOCIATED IFPROXY RPC SERVERS */
  /* Call the remote procedure "broadcastmessage" on ViNet rpc server
   * Effect: will be printed by ViNet rpc server and send to print to all IfProxy rpc servers */
  result_ptr = broadcastmessage_1(message, cl);
  if (result_ptr == NULL) {
    /* An error occurred while calling the server. Print error message and die. */
    clnt_perror(cl, server);
    exit(1);
  }
  else {
    result = *result_ptr;
  }
  /* Okay, we successfully called the remote procedure. */
  if (result != 0) {
    fprintf(stderr, "Error: server %s couldn't print your message\n", server);
    exit(1);
  }
  else {
    /* The message got printed on the server's console */ 
    printf("rpc called: broadcast message \"%s\" on %s\n", message, server);
  }
 
  /* wait on my own rpc server thread to terminate; never reached except on errors */
  pthread_join(svc_thread, (void **) result_ptr);
  /* return value of svc_thread is return value of svc_thread_main(), is pointer to a static 
     integer variable, copied into result_ptr. */
  if (*result_ptr != 0) {
    /* svc_thread returned with an error message */
    fprintf(stderr, "Error: server thread returned with error\n");
    exit(1);
  }
  
  /* never reached, as svc_thread never returns except on errors */
  return 0;
}
